/*
 *           OMEGA SECURITY SERVICES
 * 	     (c) 2008 Omega Dev Team
 *
 *
 *                 $Id: g_opm.c 1549 2009-12-13 06:54:06Z twitch $
 */


#include "stdinc.h"
#include "server.h"
#include "opm.h"
#include "opm_error.h"
#include "opm_types.h"
#include "compat.h"

#ifdef inet_pton
#undef inet_pton
#endif 

#include <arpa/inet.h>

#define NAME "g_opm"
#define AUTHOR "Omega Team"
#define VERSION "$Id: g_opm.c 1549 2009-12-13 06:54:06Z twitch $"

#define ARRAY_SIZEOF(x) (sizeof(x) / sizeof((x)[0]))

#if 0
	OPM_T *scanner;
#endif

pthread_t complete = 0;

void opm_entry(void *);
void opm_exit (void *);


void open_proxy(OPM_T *, OPM_REMOTE_T *, int, void *);
void negotiation_failed(OPM_T *, OPM_REMOTE_T *, int, void *);
void opm_timeout(OPM_T *, OPM_REMOTE_T *, int, void *);
void end(OPM_T *, OPM_REMOTE_T *, int, void *);
void handle_error(OPM_T *, OPM_REMOTE_T *, int, void *);

static void event_opmcon(int, User *);


int opm_scan_count = 0;
int opm_thread_bail = 0;

dlink_list proxy;
pthread_cond_t opm_cond;

Thread* opm_thread;


unsigned short http_ports[] = {
    8000, 8080, 3128, 80
};

unsigned short wingate_ports[] = {
    23
};

unsigned short router_ports[] = {
    23
};

unsigned short socks4_ports[] = {
    1080
};

unsigned short socks5_ports[] = {
    1080
};

unsigned short httppost_ports[] = {
    80, 8090, 3128
};

int fdlimit = 1024;
int max_read = 4096;
int port = 6667;
int err_code = 0;

int gopm_load();
void gopm_unload();
static void * gopm_signals(void * args);

MODHEADER(NAME, VERSION, AUTHOR, MAKE_ABI(0,5,2), gopm_load, gopm_unload);


int gopm_load()
{

	int status, ch;
	//if we dont have pthreads bail before we do anything. 
	#ifndef HAVE_PTHREAD
		throwModErr("Module Error, Threading is disabled or not available.","");
	#endif

	AddEvent ("USERCONN", event_opmcon);
	return MOD_CONT;
	
}




void gopm_unload()
{
	DelEvent("USERCON",event_opmcon);
	return;
}




static void event_opmcon(int ac, User *u)
{
	Thread *thread;
	unsigned char addr[16];

	if (sync_state != RUNNING)
		return;

	if (u->service)
		return;

	inet_pton(AF_INET,u->ip,&addr);

	if ((match_cidr(addr, "\x7F", 8)) && (match_cidr(addr, "\0", 8))) 
		return;
		
    if ((strcasecmp(u->ip,CfgSettings.vhost)==0) || (strcasecmp(u->ip, CfgSettings.local_ip)==0))
			return;

	if (thread = spawn_thread((void*) u, opm_entry, NULL))  
			sendto_one_notice(s_Guardian,u,"Your connection is now being scanned for open proxies.");
	else
			alog(LOG_ERROR,"Unable to spawn thread for g_opm");

	return;
}

void opm_exit (void *args)
{
        return;
}

void opm_entry(void *args)
{
	
    OPM_REMOTE_T *remote;
    OPM_T *scanner;
    OPM_ERR_T err;
	int sig;
	sigset_t set;

	char *ip;
	unsigned int i, s;
	
	int scantimeout = 15;
    time_t opm_scan_timeout =  time(NULL) + 120;

   	User *u = args;
	
	if (!u)
		return;
			
	scanner = opm_create();
	remote = opm_remote_create(u->ip);
	
	if ((!remote) || (!scanner))
		goto opm_thread_exit;
		
	sigemptyset(&set);
    sigaddset(&set, SIGPIPE);

    /* Setup callbacks */
    opm_callback(scanner, OPM_CALLBACK_OPENPROXY, &open_proxy, NULL);
    opm_callback(scanner, OPM_CALLBACK_NEGFAIL, &negotiation_failed, NULL);
    opm_callback(scanner, OPM_CALLBACK_TIMEOUT, &opm_timeout, NULL);
    opm_callback(scanner, OPM_CALLBACK_END, &end, NULL);
    opm_callback(scanner, OPM_CALLBACK_ERROR, &handle_error, NULL);


    /* Setup the scanner configuration */
    opm_config(scanner, OPM_CONFIG_FD_LIMIT, &fdlimit);
	opm_config(scanner, OPM_CONFIG_SCAN_IP, CfgSettings.local_ip);
	 
	//check for valid vhost in our config settigns
	if (CfgSettings.vhost)
		opm_config(scanner, OPM_CONFIG_BIND_IP, CfgSettings.vhost);
	else
		opm_config(scanner, OPM_CONFIG_BIND_IP, "127.0.0.1");
		
    opm_config(scanner, OPM_CONFIG_SCAN_PORT, &port);
    opm_config(scanner, OPM_CONFIG_TIMEOUT, &scantimeout);
    opm_config(scanner, OPM_CONFIG_MAX_READ, &max_read);

    /* Setup the protocol configuration */
    for (i = 0; http_ports[i]; i++)
            opm_addtype(scanner, OPM_TYPE_HTTP, http_ports[i]);


    for (i = 0; wingate_ports[i]; i++)
            opm_addtype(scanner, OPM_TYPE_WINGATE, wingate_ports[i]);


    for (i = 0; router_ports[i]; i++)
            opm_addtype(scanner, OPM_TYPE_ROUTER, router_ports[i]);


    for (i = 0; socks4_ports[i]; i++)
            opm_addtype(scanner, OPM_TYPE_SOCKS4, socks4_ports[i]);


    for (s = ARRAY_SIZEOF(socks5_ports), i = 0; i < s; i++)
            opm_addtype(scanner, OPM_TYPE_SOCKS5, socks5_ports[i]);


    for (i = 0; httppost_ports[i]; i++)
            opm_addtype(scanner, OPM_TYPE_HTTPPOST, httppost_ports[i]);
			
	    switch (err_code = opm_scan(scanner, remote))
    {
			case OPM_SUCCESS:
			//printf("Begging scan on %s\n", remote->ip);
			opm_scan_timeout = time(NULL) + 180;
			while (opm_active(scanner) > 0)
			{
				if (sigpending(&set)) {
						sigwait(&set, &sig);
				}				
				if (time(NULL) >= opm_scan_timeout)
				{
						opm_end(scanner, remote);
					break;
				}               
				opm_cycle(scanner);
				sleep(1);
			}       
			break;
			case OPM_ERR_BADADDR:
					sendto_logchan("Client connecting with a bad address (%s)", remote->ip);
				break;
			default:
					sendto_logchan("OPM Error: Unknown Error %d", err_code);
				break;
	}


	
opm_thread_exit:
	if (remote)
		opm_remote_free(remote);
	if (scanner)
		opm_free(scanner);
	dest_thread();
	return;
}


void open_proxy(OPM_T *scanner, OPM_REMOTE_T *remote, int notused, void *data)
{
   USE_VAR(notused);
   USE_VAR(data);
	//printf("Found proxy Port: %d\n",remote->port);
   sendto_logchan("\002Proxy:\002 Found proxy on %s:%d", remote->ip,
         remote->port);
    ircd_xline("Z", s_Guardian->nick, remote->ip, "Detected possible proxy from %s:%d.", remote->ip,remote->port);
    return; //meh dont continue keep matching.
}


void negotiation_failed(OPM_T *scanner, OPM_REMOTE_T *remote, int notused,
      void *data)
{
	
    USE_VAR(scanner);
    USE_VAR(notused);
    USE_VAR(data);
	//printf("Opm negotiation failed Port: %d\n",remote->port);
    return;
}

void opm_timeout(OPM_T *scanner, OPM_REMOTE_T *remote, int notused, void *data)
{
	//printf("OPM Scan timeout Port: %d\n",remote->port);
    USE_VAR(scanner);
    USE_VAR(notused);
    USE_VAR(data);	
	return;
}



void end(OPM_T *scanner, OPM_REMOTE_T *remote, int notused, void *data)
{

   USE_VAR(scanner);
   USE_VAR(notused);
   USE_VAR(data);
//printf("OPM Scan End\n");
   return;
}

void handle_error(OPM_T *scanner, OPM_REMOTE_T *remote, int err, void *data)
{

   USE_VAR(scanner);
   USE_VAR(data);
	
   switch(err)
   {
      case OPM_ERR_MAX_READ:
			 //sendto_logchan("OPM Error: Reached maxed read on %s:%d", remote->ip,
             //  remote->port);
            break;
      case OPM_ERR_BIND:
			 //sendto_logchan("OPM Error: Unable to bind to IP/Port when attempting OPM scan.");
            break;
      case OPM_ERR_NOFD:
            //sendto_logchan("OPM Error: Unable to allocate file descriptor for %s:%d",
            //   remote->ip, remote->port);
            break;
      default:
          //sendto_logchan("OPM Error: Unknown error on %s:%d, err = %d", remote->ip,
		  //remote->port, err);
			break;
   }
}


