/*
 *         	  OMEGA IRC SECURITY SERVICES
 * 	      	    (C) 2008 Omega Dev Team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *    $Id: config.c 1714 2010-01-08 06:03:04Z jer $
 */

/*
 * Dyanmic Configuration Notes:
 *
 * We have one major problem to solve regarding dynamic configurations,
 * we have to solve the chicken and the egg problem, To know what modules
 * we need to load we must first read the config, if these modules happen
 * to insert config options, amd require them to initialize we run into
 * the problem of, how do we account for these config values AFTER the
 * config has been loaded, reloading the config for each module load is
 * just a plain bad idea. I have a few theories but this defently needs
 * alot of for thought.
 *
 */

#include "stdinc.h"
#include "configparse.h"
#include "server.h"



#define PARAM_ERROR(ce, s)  { fprintf (stderr, "%s:%i: no parameter for %s " \
                                    "config option: %s\n", (ce)->ce_fileptr->cf_filename, \
                                    (ce)->ce_varlinenum, (s), (ce)->ce_varname); }


#define INVALID_PARAM(ce, s)  { fprintf (stderr, "%s:%i: Invalid parameter %s " \
                                    "config option: %s\n", (ce)->ce_fileptr->cf_filename, \
                                    (ce)->ce_varlinenum, (s), (ce)->ce_varname); \ }

#define FOREACH_PARAM(ce) for (ce = ce->ce_entries; ce; ce = ce->ce_next)

int modcnt = 0;

extern char p10_id[3];

static int  c_serverinfo        (CONFIGENTRY *);
static int  c_serverinfo_name   (CONFIGENTRY *);
static int  c_serverinfo_port   (CONFIGENTRY *);
static int  c_serverinfo_uplink (CONFIGENTRY *);
static int  c_serverinfo_desc   (CONFIGENTRY *);
static int  c_serverinfo_pass   (CONFIGENTRY *);
static int  c_serverinfo_num    (CONFIGENTRY *);
static int	c_serverinfo_sid	(CONFIGENTRY *);
static int	c_serverinfo_vhost	(CONFIGENTRY *);
static int	c_serverinfo_network(CONFIGENTRY *);

static int  c_protocol          (CONFIGENTRY *);
static int  c_protocol_name     (CONFIGENTRY *);
static int  c_protocol_ts6      (CONFIGENTRY *);


static int  c_settings				(CONFIGENTRY *);
static int  c_settings_logchan		(CONFIGENTRY *);
static int  c_settings_privmsg		(CONFIGENTRY *);
static int	c_settings_localip		(CONFIGENTRY *);
static int	c_settings_pidfile		(CONFIGENTRY *);
static int	c_settings_logusers		(CONFIGENTRY *);
static int	c_settings_lognicks		(CONFIGENTRY *);
static int  c_settings_operonly     (CONFIGENTRY *);

static int   c_guardian              (CONFIGENTRY *);
static int   c_guardian_nick         (CONFIGENTRY *);
static int   c_guardian_user         (CONFIGENTRY *);
static int   c_guardian_host         (CONFIGENTRY *);
static int   c_guardian_desc         (CONFIGENTRY *);
static int   c_guardian_channels     (CONFIGENTRY *);
static int   c_gmodules              (CONFIGENTRY *);

static int c_modules         (CONFIGENTRY *);
static int c_access         (CONFIGENTRY *);
static int c_access_flags   (CONFIGENTRY *);
static int c_access_host    (CONFIGENTRY *);


Access *CurrentAccess = NULL;

static struct ConfTable conf_root_table[] =
{
    { "SERVERINFO", 1, c_serverinfo },
    { "PROTOCOL", 1, c_protocol },
    { "SETTINGS", 1 , c_settings },
    { "GUARDIAN", 1 , c_guardian },
    { "ACCESS", 1, c_access },
	{ "MODULES", 1, c_modules },
    { NULL, 0, NULL }
};

static struct ConfTable conf_protocol_table[] =
{
    { "NAME",   0,  c_protocol_name },
    { "TS6",	0,	c_protocol_ts6 },
    { NULL,     0,  NULL }
};

static struct ConfTable conf_access_table[] =
{
	{ "FLAGS",	0,	c_access_flags },
	{ "HOSTS",  0,	c_access_host },
	{ NULL,     0,	NULL }
};

static struct ConfTable conf_settings_table[] =
{
    { "LOGCHAN",	0,  c_settings_logchan },
    { "PRIVMSG",	1,  c_settings_privmsg },
    { "LOCALIP",	1,  c_settings_localip },
    { "LOGUSERS",	0,  c_settings_logusers },
	{ "LOGNICKS",	0,  c_settings_lognicks },
	{ "PIDFILE",	0,	c_settings_pidfile },
    { "OPERONLY",   1,  c_settings_operonly },
    { NULL,     0,  NULL }
};

static struct ConfTable conf_guardian_table[] =
{
    { "NICK",               0,      c_guardian_nick },
    { "USER",               0,      c_guardian_user },
    { "HOST",               0,      c_guardian_host },
    { "DESC",               0,      c_guardian_desc },
    { "CHANNELS",           0,      c_guardian_channels },
    { "MODULES",            1,      c_gmodules },
    { NULL,                 0,      NULL }
};

static struct ConfTable conf_serverinfo_table[] =
{
    { "NAME",   0,  c_serverinfo_name },
    { "PORT",   0,  c_serverinfo_port },
    { "UPLINK", 0,  c_serverinfo_uplink },
    { "DESC",   1,  c_serverinfo_desc },
    { "PASS",   1,  c_serverinfo_pass },
    { "NUMERIC", 0, c_serverinfo_num },
    { "SID",	   0,	c_serverinfo_sid },
    { "VHOST",   0,	c_serverinfo_vhost },
	{ "NETWORK",	0,	c_serverinfo_network },
    { NULL,      0,  NULL }
};


//once done we can get rid of this god forsaken thing :)
void InitDefaults ()
{

    strcpy (CfgSettings.servername, "security.network.tld");
    strcpy (CfgSettings.uplink,		"hub.network.tld");
    strcpy (CfgSettings.pass,		"password");
    strcpy (CfgSettings.desc,		"Security Services");
    strcpy (CfgSettings.numeric,	"86");
    strcpy (CfgSettings.sid,		"666");
	strcpy (CfgSettings.network,	"Misconfigured");

    CfgSettings.port    = 6667;

	CfgSettings.vhost[0]	= '\0';

	CfgSettings.local_ip[0]	= '\0';

	snprintf (CfgSettings.pidfile, sizeof(CfgSettings.pidfile), "%s/omega.pid", DPATH);

    strcpy (logchan, "#Services");
    strcpy (guardian.nick,	"Guardian");
    strcpy (guardian.host,	"network.tld");
    strcpy (guardian.user,	"security");
    strcpy (guardian.desc,	"Network Security Service");

    CfgSettings.privmsg	    = 0;
    CfgSettings.operonly    = 1;

    memset(guardian_modules,0,sizeof(guardian_modules));
    gmod_cnt = 0;

    snprintf (datadir, sizeof(datadir), "%s", DBDIR);

}

void VerifyConf ()
{
    struct stat     buf;
    int             retval;

	if (CfgSettings.local_host != NULL)
		free (CfgSettings.local_host);

	if (CfgSettings.local_ip[0] == '\0')
		CfgSettings.local_host = get_hostname (NULL);
	else if (match_regx (CfgSettings.local_ip, "^[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}\\.[0-9]{1,3}$"))
		CfgSettings.local_host = get_hostname (NULL);
	else
		CfgSettings.local_host = get_hostname (CfgSettings.local_ip);

    if (CfgSettings.numeric != NULL)
        inttobase64 (p10_id, atoi(CfgSettings.numeric), 2); // This should be safe, numeric should be denied in rehashes

    if ((strcasecmp (CfgSettings.network, "Misconfigured")) != 0)
    {
        snprintf (datadir, sizeof(datadir), "%s/%s", DBDIR, CfgSettings.network);

        retval  = stat (datadir, &buf);

        if (retval != 0)
        {
            switch (retval)
            {
                case EACCES:
                case ELOOP:
                case ENAMETOOLONG:
                case ENOMEM:
                    fprintf (stderr, "Warning: datadir %s does not exist.\n", datadir);
                    break;
                case ENOTDIR:
                    if (mkdir (datadir, 0755) != 0)
                        fprintf (stderr, "Unable to create datadir %s: %s\n", datadir, strerror(errno));
                    break;

                default:
                    fprintf (stderr, "stat() error: %s\n", strerror(errno));

                    if (mkdir (datadir, 0755) != 0)
                        fprintf (stderr, "Unable to create datadir %s: %s\n", datadir, strerror(errno));
            }
        }
    }
    else
    {
        fprintf (stderr, "Warning: network misconfigured using %s as datadir\n", datadir);
        // Warn about using DBDIR to store databases
    }
}

void init_config() {
    dlink_node *dl;

    struct ConfTable *ct = NULL;

    for (ct = conf_root_table; ct->name; ct++)
            add_config_table(ct, CONF_REQUIRED);
}

int add_config_table(struct ConfTable *ct, int required)
{
    dlink_node *dl;

    if (!ct)
        return 0;  //null pointer passed in

    conf_entry *ce;

    if (!(ce = (conf_entry*) malloc(sizeof(conf_entry))))
        return -1; //malloc failed


    ce->ct = ct;
    ce->loaded = 0;
    ce->required = required;

    dl = dlink_create();
    dlink_add_tail(ce,dl,&config_tables);

    return 1;
}

int del_config_table(struct ConfTable *ct)
{
        dlink_node *dl,*tdl;
        struct ConfTable tmp_ct;

        if (!ct)
            return 0;

       dl = dlink_find_delete(ct,&config_tables);
       if (dl)
        dlink_free(dl);
       else
            return 0;

        return 1;
}

void ConfLoad (void)
{
    CONFIGFILE          *cfptr, *cfp;
    CONFIGENTRY         *ce;

    struct ConfTable    *ct = NULL;
    dlink_node          *dl;
    conf_entry          *conf_table;

    cfptr = cfp = config_load (CfgSettings.conf_name);

    if (cfp == NULL)
    {
        fprintf (stderr, "Unable to open configuration file (%s): %s\n",
                    CONFIG_NAME, strerror(errno));
        exit (0);
    }

    for (;cfptr;cfptr = cfptr->cf_next)
    {

        for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
        {
            DLINK_FOREACH(dl,config_tables.head)
            {
                conf_table = dl->data;

                if (!strcasecmp (conf_table->ct->name, ce->ce_varname))
                {
                    conf_table->ct->handler (ce);
                    conf_table->loaded = 1;
                    break;
                }

            }
            if (conf_table->ct->name == NULL)
                fprintf (stderr, "%s:%i invalid config option :%s\n", ce->ce_fileptr->cf_filename,
                        ce->ce_varlinenum, ce->ce_varname);
        }
    }
    config_free (cfp);

    DLINK_FOREACH(dl,config_tables.head)
    {

        conf_table = dl->data;

        if ((!conf_table->loaded) && (conf_table->required)) {
                fprintf (stderr, "Missing required config directive %s\n", conf_table->ct->name);
                exit(0);
        }
    }
}

void ConfReload (void)
{
  CONFIGFILE *cfptr;
  CONFIGFILE *cfp;
  CONFIGENTRY *ce;
  conf_entry        *conf_table;

  dlink_node *dl;
  cfptr = cfp = config_load(CfgSettings.conf_name);

  if (cfp == NULL)
  {
    fprintf (stderr, "Unable to open configuration file (%s): %s\n",
         CONFIG_NAME, strerror(errno));
  }

  for (;cfptr;cfptr = cfptr->cf_next) /* Handle include files? */
  {

    for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
    {
      DLINK_FOREACH(dl,config_tables.head)
      {
            conf_table = dl->data;
            if (!strcasecmp(conf_table->ct->name, ce->ce_varname))
            {

              if (conf_table->ct->rehashable == 1)
                    conf_table->ct->handler(ce);

              break;

            }
      }
      if (conf_table->ct->name == NULL)
		alog(LOG_ERROR, "%s:%i invalid config option: %s\n", ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
				ce->ce_varname);
    }
  }
  config_free(cfp);
}

int generic_subblock_handler(CONFIGENTRY *ce, char *subblock,
                                    struct ConfTable *table)
{
  struct ConfTable *ct = NULL;

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    for (ct = table; ct->name; ct++)
    {
        if (!strcasecmp(ct->name, ce->ce_varname))
        {
            ct->handler(ce);
            break;
        }
    }
    if (ct->name == NULL)
    {
      fprintf (stderr, "%s:%i invalid %s config option: %s\n", ce->ce_fileptr->cf_filename,
            ce->ce_varlinenum, subblock, ce->ce_varname);
    }
  }
  return 0;
}

/***************************************************/

static int c_modules(CONFIGENTRY *ce)
{
	char modname[MAXPATH];
	
   	FOREACH_PARAM(ce) {
	
		if (!ce->ce_varname)
			break;	
		snprintf(modname,sizeof(modname),"%s%s",ce->ce_varname,(strstr(ce->ce_varname,".so"))? "" : ".so");	
		addto_mod_que(modname, MOD_ACT_LOAD, MOD_LOAD_POST);
	}
	return 0;
}

/***************************************************/

static int c_gmodules (CONFIGENTRY *ce)
{
	char modname[MAXPATH];
	
   	FOREACH_PARAM(ce) {
	
		if (!ce->ce_varname)
			break;
			
		snprintf(modname,sizeof(modname),"%s%s",ce->ce_varname,(strstr(ce->ce_varname,".so"))? "" : ".so");	
		addto_mod_que(modname, MOD_ACT_LOAD, MOD_LOAD_STD);
	}
	return 0;
}

/***************************************************/

static int c_access (CONFIGENTRY *ce)
{
    FOREACH_PARAM(ce)
    {
		if (!ce->ce_varname)
				continue;
				
		//the only reason we return null here is if we cant malloc
		if (!(CurrentAccess = new_access(ce->ce_varname)))
				continue;
				
		generic_subblock_handler (ce, "ACCESS", conf_access_table);
    }
    return 0;
}

/***************************************************/

static int c_access_host (CONFIGENTRY *ce)
{
	char *tmp_host;
	dlink_node *dl;
	int rtn = 0;
	
	FOREACH_PARAM(ce)		
            add_access_host(CurrentAccess, ce->ce_varname);
			
	return 0;
}

/***************************************************/

static int c_access_flags (CONFIGENTRY *ce)
{
	FOREACH_PARAM(ce)
	{
		if (ce->ce_varname)
			set_access_flag(CurrentAccess, ce->ce_varname);
	}
	return 0;
}


/***************************************************/
static int c_serverinfo (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "SERVERINFO", conf_serverinfo_table);
    return 0;
}



static int c_serverinfo_num (CONFIGENTRY *ce)
{
   if (ce->ce_vardata == NULL)
   {
        PARAM_ERROR (ce,"serverinfo");
   }

   strlcpy(CfgSettings.numeric,ce->ce_vardata, sizeof(CfgSettings.numeric));
   return 0;
}

static int c_serverinfo_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "serverinfo");
    }

    strlcpy (CfgSettings.servername, ce->ce_vardata, sizeof(CfgSettings.servername));
    return 0;
}

static int c_serverinfo_port (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "serverinfo");
    }

    CfgSettings.port    = ce->ce_vardatanum;
    return 0;
}

static int c_serverinfo_uplink (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "serverinfo");
    }

    /* Due to a bug I found awhile back, where IPv6 can be enclapsed in [], we must account for this */

    if (ce->ce_vardata[0] == '[')
    {
        int len;
        ce->ce_vardata++;
        len = strlen(ce->ce_vardata);

        if ((len > 0) && (ce->ce_vardata[len - 1] == ']'))
            ce->ce_vardata[len - 1] = '\0';

        fprintf (stderr, "%s:%i WARNING: Brackets [] used in serverinfo uplink", ce->ce_fileptr->cf_filename,
                ce->ce_varlinenum);
    }

    strlcpy (CfgSettings.uplink, ce->ce_vardata, sizeof(CfgSettings.uplink));
    return 0;
}

static int c_serverinfo_pass (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "serverinfo");
    }

    strlcpy (CfgSettings.pass, ce->ce_vardata, sizeof(CfgSettings.pass));
    return 0;
}

static int c_serverinfo_desc (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "serverinfo");
    }

    strlcpy (CfgSettings.desc, ce->ce_vardata, sizeof(CfgSettings.desc));
    return 0;
}

static int c_serverinfo_sid (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "serverinfo");
	}

	strlcpy (CfgSettings.sid, ce->ce_vardata, sizeof(CfgSettings.sid));
	return 0;
}

static int c_serverinfo_vhost (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "serverinfo");
	}

	if (ce->ce_vardata[0] == '[')
	{
		int	len;
		ce->ce_vardata++;

		len	= strlen(ce->ce_vardata);

		if ((len > 0) && (ce->ce_vardata[len - 1] == ']'))
			ce->ce_vardata[len - 1] = '\0';
	}

	strlcpy (CfgSettings.vhost, ce->ce_vardata, sizeof(CfgSettings.vhost));

	return 0;
}

static int c_serverinfo_network (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "serverinfo");
	}

	strlcpy (CfgSettings.network, ce->ce_vardata, sizeof(CfgSettings.network));
	return 0;
}

/*********************************************/

static int c_protocol (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "PROTOCOL", conf_protocol_table);
    return 0;
}


static int c_protocol_name (CONFIGENTRY *ce)
{


    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "protocol");
    }

    strlcpy (CfgSettings.protocol, ce->ce_vardata, sizeof(CfgSettings.protocol));
     return 0;
}

static int c_protocol_ts6 (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "protocol");
	}

	IRCd.ts6	= ce->ce_vardatanum;
	return 0;
}

/*******************************************/

static int c_settings (CONFIGENTRY *ce)
{
    logclients = 0; //default this to 0 so if were rehashed
                    //and its commented we will know :)

    generic_subblock_handler (ce, "SETTINGS", conf_settings_table);
    return 0;
}


static int c_settings_logchan (CONFIGENTRY *ce)
{

    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "settings");
    }

    strlcpy (logchan, ce->ce_vardata, sizeof(logchan));
    return 0;
}

static int c_settings_privmsg (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "settings");
    }

    CfgSettings.privmsg = ce->ce_vardatanum;
    return 0;
}

static int c_settings_localip (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "settings");
	}

	strlcpy (CfgSettings.local_ip, ce->ce_vardata, sizeof(CfgSettings.local_ip));
	return 0;
}

static int c_settings_pidfile (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "settings");
	}

	snprintf (CfgSettings.pidfile, sizeof(CfgSettings.pidfile), "%s/%s", DPATH, ce->ce_vardata);
	return 0;
}

static int c_settings_lognicks (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "settings");
	}

	lognicks	= ce->ce_vardatanum;
	return 0;
}

static int c_settings_logusers (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "settings");
	}

	logclients	= ce->ce_vardatanum;
	return 0;
}

static int c_settings_operonly (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "settings");
    }

    CfgSettings.operonly = ce->ce_vardatanum;
    return 0;
}

/*******************************************/


static int c_guardian (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "GUARDIAN", conf_guardian_table);
    return 0;
}

static int c_guardian_nick (CONFIGENTRY *ce)
{
        if (ce->ce_vardata == NULL)
        {
                PARAM_ERROR (ce, "guardian");
        }

    strlcpy (guardian.nick, ce->ce_vardata, sizeof(guardian.nick));
        return 0;
}


static int c_guardian_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "guardian");
    }

    strlcpy (guardian.user, ce->ce_vardata, sizeof(guardian.user));
    return 0;
}

static int c_guardian_host (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "guardian");
    }

    strlcpy (guardian.host, ce->ce_vardata, sizeof(guardian.host));
    return 0;
}

static int c_guardian_desc (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "guardian");
    }

    strlcpy (guardian.desc, ce->ce_vardata, sizeof(guardian.desc));
    return 0;
}


static int c_guardian_channels (CONFIGENTRY *ce)
{
	char	*chan;
	dlink_node	*dl;

        if (ce->ce_varname == NULL)
        {
                PARAM_ERROR (ce, "guardian");
        }

        for (ce = ce->ce_entries;ce;ce = ce->ce_next)
        {
			chan	= malloc(sizeof(chan));
			dl		= dlink_create ();

			chan	= strdup(ce->ce_varname);
			dlink_add	(chan, dl, &guardian.chanlist);
        }

        return 0;
}

